# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: aibsmw_messages.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='aibsmw_messages.proto',
  package='',
  syntax='proto2',
  serialized_pb=_b('\n\x15\x61ibsmw_messages.proto\"V\n\x0emessage_header\x12\x0f\n\x07process\x18\x01 \x02(\t\x12\x0c\n\x04host\x18\x02 \x02(\t\x12\x11\n\ttimestamp\x18\x03 \x02(\x02\x12\x12\n\nmessage_id\x18\x04 \x02(\t\"H\n\x11generic_heartbeat\x12\x1f\n\x06header\x18\x01 \x02(\x0b\x32\x0f.message_header\x12\x12\n\nstart_time\x18\x02 \x02(\x02\"\x85\x01\n\x17remote_device_heartbeat\x12\x1f\n\x06header\x18\x01 \x02(\x0b\x32\x0f.message_header\x12\x13\n\x0b\x64\x65vice_name\x18\x02 \x02(\t\x12\x12\n\nip_address\x18\x04 \x02(\t\x12\x0c\n\x04port\x18\x05 \x02(\x05\x12\x12\n\nstart_time\x18\x06 \x02(\x02\"9\n\x16request_remote_devices\x12\x1f\n\x06header\x18\x01 \x02(\x0b\x32\x0f.message_header\"L\n\x0crouter_alive\x12\x1f\n\x06header\x18\x01 \x02(\x0b\x32\x0f.message_header\x12\x1b\n\x13registered_messages\x18\x02 \x03(\t\"G\n\x13remote_devices_list\x12\x1f\n\x06header\x18\x01 \x02(\x0b\x32\x0f.message_header\x12\x0f\n\x07\x64\x65vices\x18\x02 \x02(\t\"\x85\x02\n\x16remote_service_request\x12\x1f\n\x06header\x18\x01 \x02(\x0b\x32\x0f.message_header\x12;\n\x0c\x63ommand_type\x18\x02 \x02(\x0e\x32%.remote_service_request.COMMAND_TYPES\x12\x0e\n\x06target\x18\x03 \x02(\t\x12\x0c\n\x04\x61rgs\x18\x04 \x01(\t\x12\x0e\n\x06kwargs\x18\x05 \x01(\t\"_\n\rCOMMAND_TYPES\x12\x0b\n\x07\x43MD_RUN\x10\x00\x12\x0b\n\x07\x43MD_SET\x10\x01\x12\x0b\n\x07\x43MD_GET\x10\x02\x12\x10\n\x0c\x43MD_CALLABLE\x10\x03\x12\x15\n\x11\x43MD_PLATFORM_INFO\x10\x04\"\xc2\x01\n\x14remote_service_reply\x12\x1f\n\x06header\x18\x01 \x02(\x0b\x32\x0f.message_header\x12<\n\x0b\x63\x61ll_result\x18\x02 \x02(\x0e\x32\'.remote_service_reply.CALL_RESULT_TYPES\x12\r\n\x05reply\x18\x03 \x02(\t\"<\n\x11\x43\x41LL_RESULT_TYPES\x12\x11\n\rRESULT_FAILED\x10\x00\x12\x14\n\x10RESULT_PROCESSED\x10\x01\"\xbb\x01\n\x0bpython_info\x12\x14\n\x0c\x62uild_number\x18\x01 \x02(\t\x12\x12\n\nbuild_date\x18\x02 \x02(\t\x12\x10\n\x08\x63ompiler\x18\x03 \x02(\t\x12\x0e\n\x06\x62ranch\x18\x04 \x02(\t\x12\x16\n\x0eimplementation\x18\x05 \x02(\t\x12\x10\n\x08revision\x18\x06 \x02(\t\x12\x0f\n\x07version\x18\x07 \x02(\t\x12\x13\n\x0b\x65xec_prefix\x18\x08 \x02(\t\x12\x10\n\x08is_conda\x18\n \x02(\x08\"\xaa\x01\n\thost_info\x12\x0f\n\x07machine\x18\x01 \x02(\t\x12\x0c\n\x04node\x18\x02 \x02(\t\x12\x10\n\x08platform\x18\x03 \x02(\t\x12\x11\n\tprocessor\x18\x04 \x02(\t\x12\x0f\n\x07release\x18\x05 \x02(\t\x12\x0e\n\x06system\x18\x06 \x02(\t\x12\x0f\n\x07version\x18\x07 \x02(\t\x12\x14\n\x0csys_platform\x18\x08 \x02(\t\x12\x11\n\tbyteorder\x18\t \x02(\t\"|\n\rplatform_info\x12\x1f\n\x06header\x18\x01 \x02(\x0b\x32\x0f.message_header\x12\x1c\n\x06python\x18\x02 \x02(\x0b\x32\x0c.python_info\x12\x18\n\x04host\x18\x03 \x02(\x0b\x32\n.host_info\x12\x12\n\nstart_time\x18\x04 \x02(\x02\"K\n\x14register_for_message\x12\x1f\n\x06header\x18\x01 \x02(\x0b\x32\x0f.message_header\x12\x12\n\nmessage_id\x18\x02 \x02(\t\"M\n\x16\x64\x65register_for_message\x12\x1f\n\x06header\x18\x01 \x02(\x0b\x32\x0f.message_header\x12\x12\n\nmessage_id\x18\x02 \x02(\t')
)



_REMOTE_SERVICE_REQUEST_COMMAND_TYPES = _descriptor.EnumDescriptor(
  name='COMMAND_TYPES',
  full_name='remote_service_request.COMMAND_TYPES',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='CMD_RUN', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CMD_SET', index=1, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CMD_GET', index=2, number=2,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CMD_CALLABLE', index=3, number=3,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CMD_PLATFORM_INFO', index=4, number=4,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=700,
  serialized_end=795,
)
_sym_db.RegisterEnumDescriptor(_REMOTE_SERVICE_REQUEST_COMMAND_TYPES)

_REMOTE_SERVICE_REPLY_CALL_RESULT_TYPES = _descriptor.EnumDescriptor(
  name='CALL_RESULT_TYPES',
  full_name='remote_service_reply.CALL_RESULT_TYPES',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='RESULT_FAILED', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RESULT_PROCESSED', index=1, number=1,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=932,
  serialized_end=992,
)
_sym_db.RegisterEnumDescriptor(_REMOTE_SERVICE_REPLY_CALL_RESULT_TYPES)


_MESSAGE_HEADER = _descriptor.Descriptor(
  name='message_header',
  full_name='message_header',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='process', full_name='message_header.process', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='host', full_name='message_header.host', index=1,
      number=2, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='timestamp', full_name='message_header.timestamp', index=2,
      number=3, type=2, cpp_type=6, label=2,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='message_id', full_name='message_header.message_id', index=3,
      number=4, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=25,
  serialized_end=111,
)


_GENERIC_HEARTBEAT = _descriptor.Descriptor(
  name='generic_heartbeat',
  full_name='generic_heartbeat',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='header', full_name='generic_heartbeat.header', index=0,
      number=1, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='start_time', full_name='generic_heartbeat.start_time', index=1,
      number=2, type=2, cpp_type=6, label=2,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=113,
  serialized_end=185,
)


_REMOTE_DEVICE_HEARTBEAT = _descriptor.Descriptor(
  name='remote_device_heartbeat',
  full_name='remote_device_heartbeat',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='header', full_name='remote_device_heartbeat.header', index=0,
      number=1, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='device_name', full_name='remote_device_heartbeat.device_name', index=1,
      number=2, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='ip_address', full_name='remote_device_heartbeat.ip_address', index=2,
      number=4, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='port', full_name='remote_device_heartbeat.port', index=3,
      number=5, type=5, cpp_type=1, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='start_time', full_name='remote_device_heartbeat.start_time', index=4,
      number=6, type=2, cpp_type=6, label=2,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=188,
  serialized_end=321,
)


_REQUEST_REMOTE_DEVICES = _descriptor.Descriptor(
  name='request_remote_devices',
  full_name='request_remote_devices',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='header', full_name='request_remote_devices.header', index=0,
      number=1, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=323,
  serialized_end=380,
)


_ROUTER_ALIVE = _descriptor.Descriptor(
  name='router_alive',
  full_name='router_alive',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='header', full_name='router_alive.header', index=0,
      number=1, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='registered_messages', full_name='router_alive.registered_messages', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=382,
  serialized_end=458,
)


_REMOTE_DEVICES_LIST = _descriptor.Descriptor(
  name='remote_devices_list',
  full_name='remote_devices_list',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='header', full_name='remote_devices_list.header', index=0,
      number=1, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='devices', full_name='remote_devices_list.devices', index=1,
      number=2, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=460,
  serialized_end=531,
)


_REMOTE_SERVICE_REQUEST = _descriptor.Descriptor(
  name='remote_service_request',
  full_name='remote_service_request',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='header', full_name='remote_service_request.header', index=0,
      number=1, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='command_type', full_name='remote_service_request.command_type', index=1,
      number=2, type=14, cpp_type=8, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='target', full_name='remote_service_request.target', index=2,
      number=3, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='args', full_name='remote_service_request.args', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='kwargs', full_name='remote_service_request.kwargs', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _REMOTE_SERVICE_REQUEST_COMMAND_TYPES,
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=534,
  serialized_end=795,
)


_REMOTE_SERVICE_REPLY = _descriptor.Descriptor(
  name='remote_service_reply',
  full_name='remote_service_reply',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='header', full_name='remote_service_reply.header', index=0,
      number=1, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='call_result', full_name='remote_service_reply.call_result', index=1,
      number=2, type=14, cpp_type=8, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='reply', full_name='remote_service_reply.reply', index=2,
      number=3, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _REMOTE_SERVICE_REPLY_CALL_RESULT_TYPES,
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=798,
  serialized_end=992,
)


_PYTHON_INFO = _descriptor.Descriptor(
  name='python_info',
  full_name='python_info',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='build_number', full_name='python_info.build_number', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='build_date', full_name='python_info.build_date', index=1,
      number=2, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='compiler', full_name='python_info.compiler', index=2,
      number=3, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='branch', full_name='python_info.branch', index=3,
      number=4, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='implementation', full_name='python_info.implementation', index=4,
      number=5, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='revision', full_name='python_info.revision', index=5,
      number=6, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='version', full_name='python_info.version', index=6,
      number=7, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='exec_prefix', full_name='python_info.exec_prefix', index=7,
      number=8, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='is_conda', full_name='python_info.is_conda', index=8,
      number=10, type=8, cpp_type=7, label=2,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=995,
  serialized_end=1182,
)


_HOST_INFO = _descriptor.Descriptor(
  name='host_info',
  full_name='host_info',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='machine', full_name='host_info.machine', index=0,
      number=1, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='node', full_name='host_info.node', index=1,
      number=2, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='platform', full_name='host_info.platform', index=2,
      number=3, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='processor', full_name='host_info.processor', index=3,
      number=4, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='release', full_name='host_info.release', index=4,
      number=5, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='system', full_name='host_info.system', index=5,
      number=6, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='version', full_name='host_info.version', index=6,
      number=7, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='sys_platform', full_name='host_info.sys_platform', index=7,
      number=8, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='byteorder', full_name='host_info.byteorder', index=8,
      number=9, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1185,
  serialized_end=1355,
)


_PLATFORM_INFO = _descriptor.Descriptor(
  name='platform_info',
  full_name='platform_info',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='header', full_name='platform_info.header', index=0,
      number=1, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='python', full_name='platform_info.python', index=1,
      number=2, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='host', full_name='platform_info.host', index=2,
      number=3, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='start_time', full_name='platform_info.start_time', index=3,
      number=4, type=2, cpp_type=6, label=2,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1357,
  serialized_end=1481,
)


_REGISTER_FOR_MESSAGE = _descriptor.Descriptor(
  name='register_for_message',
  full_name='register_for_message',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='header', full_name='register_for_message.header', index=0,
      number=1, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='message_id', full_name='register_for_message.message_id', index=1,
      number=2, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1483,
  serialized_end=1558,
)


_DEREGISTER_FOR_MESSAGE = _descriptor.Descriptor(
  name='deregister_for_message',
  full_name='deregister_for_message',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='header', full_name='deregister_for_message.header', index=0,
      number=1, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='message_id', full_name='deregister_for_message.message_id', index=1,
      number=2, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1560,
  serialized_end=1637,
)

_GENERIC_HEARTBEAT.fields_by_name['header'].message_type = _MESSAGE_HEADER
_REMOTE_DEVICE_HEARTBEAT.fields_by_name['header'].message_type = _MESSAGE_HEADER
_REQUEST_REMOTE_DEVICES.fields_by_name['header'].message_type = _MESSAGE_HEADER
_ROUTER_ALIVE.fields_by_name['header'].message_type = _MESSAGE_HEADER
_REMOTE_DEVICES_LIST.fields_by_name['header'].message_type = _MESSAGE_HEADER
_REMOTE_SERVICE_REQUEST.fields_by_name['header'].message_type = _MESSAGE_HEADER
_REMOTE_SERVICE_REQUEST.fields_by_name['command_type'].enum_type = _REMOTE_SERVICE_REQUEST_COMMAND_TYPES
_REMOTE_SERVICE_REQUEST_COMMAND_TYPES.containing_type = _REMOTE_SERVICE_REQUEST
_REMOTE_SERVICE_REPLY.fields_by_name['header'].message_type = _MESSAGE_HEADER
_REMOTE_SERVICE_REPLY.fields_by_name['call_result'].enum_type = _REMOTE_SERVICE_REPLY_CALL_RESULT_TYPES
_REMOTE_SERVICE_REPLY_CALL_RESULT_TYPES.containing_type = _REMOTE_SERVICE_REPLY
_PLATFORM_INFO.fields_by_name['header'].message_type = _MESSAGE_HEADER
_PLATFORM_INFO.fields_by_name['python'].message_type = _PYTHON_INFO
_PLATFORM_INFO.fields_by_name['host'].message_type = _HOST_INFO
_REGISTER_FOR_MESSAGE.fields_by_name['header'].message_type = _MESSAGE_HEADER
_DEREGISTER_FOR_MESSAGE.fields_by_name['header'].message_type = _MESSAGE_HEADER
DESCRIPTOR.message_types_by_name['message_header'] = _MESSAGE_HEADER
DESCRIPTOR.message_types_by_name['generic_heartbeat'] = _GENERIC_HEARTBEAT
DESCRIPTOR.message_types_by_name['remote_device_heartbeat'] = _REMOTE_DEVICE_HEARTBEAT
DESCRIPTOR.message_types_by_name['request_remote_devices'] = _REQUEST_REMOTE_DEVICES
DESCRIPTOR.message_types_by_name['router_alive'] = _ROUTER_ALIVE
DESCRIPTOR.message_types_by_name['remote_devices_list'] = _REMOTE_DEVICES_LIST
DESCRIPTOR.message_types_by_name['remote_service_request'] = _REMOTE_SERVICE_REQUEST
DESCRIPTOR.message_types_by_name['remote_service_reply'] = _REMOTE_SERVICE_REPLY
DESCRIPTOR.message_types_by_name['python_info'] = _PYTHON_INFO
DESCRIPTOR.message_types_by_name['host_info'] = _HOST_INFO
DESCRIPTOR.message_types_by_name['platform_info'] = _PLATFORM_INFO
DESCRIPTOR.message_types_by_name['register_for_message'] = _REGISTER_FOR_MESSAGE
DESCRIPTOR.message_types_by_name['deregister_for_message'] = _DEREGISTER_FOR_MESSAGE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

message_header = _reflection.GeneratedProtocolMessageType('message_header', (_message.Message,), dict(
  DESCRIPTOR = _MESSAGE_HEADER,
  __module__ = 'aibsmw_messages_pb2'
  # @@protoc_insertion_point(class_scope:message_header)
  ))
_sym_db.RegisterMessage(message_header)

generic_heartbeat = _reflection.GeneratedProtocolMessageType('generic_heartbeat', (_message.Message,), dict(
  DESCRIPTOR = _GENERIC_HEARTBEAT,
  __module__ = 'aibsmw_messages_pb2'
  # @@protoc_insertion_point(class_scope:generic_heartbeat)
  ))
_sym_db.RegisterMessage(generic_heartbeat)

remote_device_heartbeat = _reflection.GeneratedProtocolMessageType('remote_device_heartbeat', (_message.Message,), dict(
  DESCRIPTOR = _REMOTE_DEVICE_HEARTBEAT,
  __module__ = 'aibsmw_messages_pb2'
  # @@protoc_insertion_point(class_scope:remote_device_heartbeat)
  ))
_sym_db.RegisterMessage(remote_device_heartbeat)

request_remote_devices = _reflection.GeneratedProtocolMessageType('request_remote_devices', (_message.Message,), dict(
  DESCRIPTOR = _REQUEST_REMOTE_DEVICES,
  __module__ = 'aibsmw_messages_pb2'
  # @@protoc_insertion_point(class_scope:request_remote_devices)
  ))
_sym_db.RegisterMessage(request_remote_devices)

router_alive = _reflection.GeneratedProtocolMessageType('router_alive', (_message.Message,), dict(
  DESCRIPTOR = _ROUTER_ALIVE,
  __module__ = 'aibsmw_messages_pb2'
  # @@protoc_insertion_point(class_scope:router_alive)
  ))
_sym_db.RegisterMessage(router_alive)

remote_devices_list = _reflection.GeneratedProtocolMessageType('remote_devices_list', (_message.Message,), dict(
  DESCRIPTOR = _REMOTE_DEVICES_LIST,
  __module__ = 'aibsmw_messages_pb2'
  # @@protoc_insertion_point(class_scope:remote_devices_list)
  ))
_sym_db.RegisterMessage(remote_devices_list)

remote_service_request = _reflection.GeneratedProtocolMessageType('remote_service_request', (_message.Message,), dict(
  DESCRIPTOR = _REMOTE_SERVICE_REQUEST,
  __module__ = 'aibsmw_messages_pb2'
  # @@protoc_insertion_point(class_scope:remote_service_request)
  ))
_sym_db.RegisterMessage(remote_service_request)

remote_service_reply = _reflection.GeneratedProtocolMessageType('remote_service_reply', (_message.Message,), dict(
  DESCRIPTOR = _REMOTE_SERVICE_REPLY,
  __module__ = 'aibsmw_messages_pb2'
  # @@protoc_insertion_point(class_scope:remote_service_reply)
  ))
_sym_db.RegisterMessage(remote_service_reply)

python_info = _reflection.GeneratedProtocolMessageType('python_info', (_message.Message,), dict(
  DESCRIPTOR = _PYTHON_INFO,
  __module__ = 'aibsmw_messages_pb2'
  # @@protoc_insertion_point(class_scope:python_info)
  ))
_sym_db.RegisterMessage(python_info)

host_info = _reflection.GeneratedProtocolMessageType('host_info', (_message.Message,), dict(
  DESCRIPTOR = _HOST_INFO,
  __module__ = 'aibsmw_messages_pb2'
  # @@protoc_insertion_point(class_scope:host_info)
  ))
_sym_db.RegisterMessage(host_info)

platform_info = _reflection.GeneratedProtocolMessageType('platform_info', (_message.Message,), dict(
  DESCRIPTOR = _PLATFORM_INFO,
  __module__ = 'aibsmw_messages_pb2'
  # @@protoc_insertion_point(class_scope:platform_info)
  ))
_sym_db.RegisterMessage(platform_info)

register_for_message = _reflection.GeneratedProtocolMessageType('register_for_message', (_message.Message,), dict(
  DESCRIPTOR = _REGISTER_FOR_MESSAGE,
  __module__ = 'aibsmw_messages_pb2'
  # @@protoc_insertion_point(class_scope:register_for_message)
  ))
_sym_db.RegisterMessage(register_for_message)

deregister_for_message = _reflection.GeneratedProtocolMessageType('deregister_for_message', (_message.Message,), dict(
  DESCRIPTOR = _DEREGISTER_FOR_MESSAGE,
  __module__ = 'aibsmw_messages_pb2'
  # @@protoc_insertion_point(class_scope:deregister_for_message)
  ))
_sym_db.RegisterMessage(deregister_for_message)


# @@protoc_insertion_point(module_scope)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: middleware_messages.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import aibsmw_messages_pb2 as aibsmw__messages__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='middleware_messages.proto',
  package='',
  syntax='proto2',
  serialized_pb=_b('\n\x19middleware_messages.proto\x1a\x15\x61ibsmw_messages.proto\"=\n\x08workflow\x12\x1f\n\x06header\x18\x01 \x02(\x0b\x32\x0f.message_header\x12\x10\n\x08workflow\x18\x02 \x02(\t\"3\n\x10workflow_request\x12\x1f\n\x06header\x18\x01 \x02(\x0b\x32\x0f.message_header\"H\n\rworkflow_list\x12\x1f\n\x06header\x18\x01 \x02(\x0b\x32\x0f.message_header\x12\x16\n\x0eworkflow_files\x18\x02 \x03(\t\"f\n\rworkflow_load\x12\x1f\n\x06header\x18\x01 \x02(\x0b\x32\x0f.message_header\x12\x19\n\x11workflow_filename\x18\x02 \x02(\t\x12\x19\n\ndummy_mode\x18\x03 \x02(\x08:\x05\x66\x61lse\"8\n\x15workflow_list_request\x12\x1f\n\x06header\x18\x01 \x02(\x0b\x32\x0f.message_header\"L\n\x15workflow_revert_state\x12\x1f\n\x06header\x18\x01 \x02(\x0b\x32\x0f.message_header\x12\x12\n\nstate_name\x18\x02 \x02(\t\"`\n\x13transition_to_state\x12\x1f\n\x06header\x18\x01 \x02(\x0b\x32\x0f.message_header\x12\x12\n\nstate_name\x18\x02 \x02(\t\x12\x14\n\x0cstate_object\x18\x03 \x02(\t\"b\n\x15transition_to_cleanup\x12\x1f\n\x06header\x18\x01 \x02(\x0b\x32\x0f.message_header\x12\x12\n\nstate_name\x18\x02 \x02(\t\x12\x14\n\x0cstate_object\x18\x03 \x02(\t\"\xc6\x01\n\rstate_changed\x12\x1f\n\x06header\x18\x01 \x02(\x0b\x32\x0f.message_header\x12\x12\n\nstate_name\x18\x02 \x02(\t\x12+\n\x06result\x18\x03 \x02(\x0e\x32\x1b.state_changed.result_types\x12\x16\n\x0estatus_message\x18\x04 \x02(\t\x12\x14\n\x0cstate_object\x18\x05 \x02(\t\"%\n\x0cresult_types\x12\x0b\n\x07SUCCESS\x10\x00\x12\x08\n\x04\x46\x41IL\x10\x01\">\n\nstate_busy\x12\x1f\n\x06header\x18\x01 \x02(\x0b\x32\x0f.message_header\x12\x0f\n\x07message\x18\x02 \x02(\t\"?\n\x0bstate_ready\x12\x1f\n\x06header\x18\x01 \x02(\x0b\x32\x0f.message_header\x12\x0f\n\x07message\x18\x02 \x02(\t\"7\n\x03log\x12\x1f\n\x06header\x18\x01 \x02(\x0b\x32\x0f.message_header\x12\x0f\n\x07message\x18\x02 \x02(\t')
  ,
  dependencies=[aibsmw__messages__pb2.DESCRIPTOR,])
_sym_db.RegisterFileDescriptor(DESCRIPTOR)



_STATE_CHANGED_RESULT_TYPES = _descriptor.EnumDescriptor(
  name='result_types',
  full_name='state_changed.result_types',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='SUCCESS', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FAIL', index=1, number=1,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=842,
  serialized_end=879,
)
_sym_db.RegisterEnumDescriptor(_STATE_CHANGED_RESULT_TYPES)


_WORKFLOW = _descriptor.Descriptor(
  name='workflow',
  full_name='workflow',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='header', full_name='workflow.header', index=0,
      number=1, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='workflow', full_name='workflow.workflow', index=1,
      number=2, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=52,
  serialized_end=113,
)


_WORKFLOW_REQUEST = _descriptor.Descriptor(
  name='workflow_request',
  full_name='workflow_request',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='header', full_name='workflow_request.header', index=0,
      number=1, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=115,
  serialized_end=166,
)


_WORKFLOW_LIST = _descriptor.Descriptor(
  name='workflow_list',
  full_name='workflow_list',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='header', full_name='workflow_list.header', index=0,
      number=1, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='workflow_files', full_name='workflow_list.workflow_files', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=168,
  serialized_end=240,
)


_WORKFLOW_LOAD = _descriptor.Descriptor(
  name='workflow_load',
  full_name='workflow_load',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='header', full_name='workflow_load.header', index=0,
      number=1, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='workflow_filename', full_name='workflow_load.workflow_filename', index=1,
      number=2, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='dummy_mode', full_name='workflow_load.dummy_mode', index=2,
      number=3, type=8, cpp_type=7, label=2,
      has_default_value=True, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=242,
  serialized_end=344,
)


_WORKFLOW_LIST_REQUEST = _descriptor.Descriptor(
  name='workflow_list_request',
  full_name='workflow_list_request',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='header', full_name='workflow_list_request.header', index=0,
      number=1, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=346,
  serialized_end=402,
)


_WORKFLOW_REVERT_STATE = _descriptor.Descriptor(
  name='workflow_revert_state',
  full_name='workflow_revert_state',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='header', full_name='workflow_revert_state.header', index=0,
      number=1, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='state_name', full_name='workflow_revert_state.state_name', index=1,
      number=2, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=404,
  serialized_end=480,
)


_TRANSITION_TO_STATE = _descriptor.Descriptor(
  name='transition_to_state',
  full_name='transition_to_state',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='header', full_name='transition_to_state.header', index=0,
      number=1, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='state_name', full_name='transition_to_state.state_name', index=1,
      number=2, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='state_object', full_name='transition_to_state.state_object', index=2,
      number=3, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=482,
  serialized_end=578,
)


_TRANSITION_TO_CLEANUP = _descriptor.Descriptor(
  name='transition_to_cleanup',
  full_name='transition_to_cleanup',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='header', full_name='transition_to_cleanup.header', index=0,
      number=1, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='state_name', full_name='transition_to_cleanup.state_name', index=1,
      number=2, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='state_object', full_name='transition_to_cleanup.state_object', index=2,
      number=3, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=580,
  serialized_end=678,
)


_STATE_CHANGED = _descriptor.Descriptor(
  name='state_changed',
  full_name='state_changed',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='header', full_name='state_changed.header', index=0,
      number=1, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='state_name', full_name='state_changed.state_name', index=1,
      number=2, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='result', full_name='state_changed.result', index=2,
      number=3, type=14, cpp_type=8, label=2,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='status_message', full_name='state_changed.status_message', index=3,
      number=4, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='state_object', full_name='state_changed.state_object', index=4,
      number=5, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _STATE_CHANGED_RESULT_TYPES,
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=681,
  serialized_end=879,
)


_STATE_BUSY = _descriptor.Descriptor(
  name='state_busy',
  full_name='state_busy',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='header', full_name='state_busy.header', index=0,
      number=1, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='message', full_name='state_busy.message', index=1,
      number=2, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=881,
  serialized_end=943,
)


_STATE_READY = _descriptor.Descriptor(
  name='state_ready',
  full_name='state_ready',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='header', full_name='state_ready.header', index=0,
      number=1, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='message', full_name='state_ready.message', index=1,
      number=2, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=945,
  serialized_end=1008,
)


_LOG = _descriptor.Descriptor(
  name='log',
  full_name='log',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='header', full_name='log.header', index=0,
      number=1, type=11, cpp_type=10, label=2,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='message', full_name='log.message', index=1,
      number=2, type=9, cpp_type=9, label=2,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1010,
  serialized_end=1065,
)

_WORKFLOW.fields_by_name['header'].message_type = aibsmw__messages__pb2._MESSAGE_HEADER
_WORKFLOW_REQUEST.fields_by_name['header'].message_type = aibsmw__messages__pb2._MESSAGE_HEADER
_WORKFLOW_LIST.fields_by_name['header'].message_type = aibsmw__messages__pb2._MESSAGE_HEADER
_WORKFLOW_LOAD.fields_by_name['header'].message_type = aibsmw__messages__pb2._MESSAGE_HEADER
_WORKFLOW_LIST_REQUEST.fields_by_name['header'].message_type = aibsmw__messages__pb2._MESSAGE_HEADER
_WORKFLOW_REVERT_STATE.fields_by_name['header'].message_type = aibsmw__messages__pb2._MESSAGE_HEADER
_TRANSITION_TO_STATE.fields_by_name['header'].message_type = aibsmw__messages__pb2._MESSAGE_HEADER
_TRANSITION_TO_CLEANUP.fields_by_name['header'].message_type = aibsmw__messages__pb2._MESSAGE_HEADER
_STATE_CHANGED.fields_by_name['header'].message_type = aibsmw__messages__pb2._MESSAGE_HEADER
_STATE_CHANGED.fields_by_name['result'].enum_type = _STATE_CHANGED_RESULT_TYPES
_STATE_CHANGED_RESULT_TYPES.containing_type = _STATE_CHANGED
_STATE_BUSY.fields_by_name['header'].message_type = aibsmw__messages__pb2._MESSAGE_HEADER
_STATE_READY.fields_by_name['header'].message_type = aibsmw__messages__pb2._MESSAGE_HEADER
_LOG.fields_by_name['header'].message_type = aibsmw__messages__pb2._MESSAGE_HEADER
DESCRIPTOR.message_types_by_name['workflow'] = _WORKFLOW
DESCRIPTOR.message_types_by_name['workflow_request'] = _WORKFLOW_REQUEST
DESCRIPTOR.message_types_by_name['workflow_list'] = _WORKFLOW_LIST
DESCRIPTOR.message_types_by_name['workflow_load'] = _WORKFLOW_LOAD
DESCRIPTOR.message_types_by_name['workflow_list_request'] = _WORKFLOW_LIST_REQUEST
DESCRIPTOR.message_types_by_name['workflow_revert_state'] = _WORKFLOW_REVERT_STATE
DESCRIPTOR.message_types_by_name['transition_to_state'] = _TRANSITION_TO_STATE
DESCRIPTOR.message_types_by_name['transition_to_cleanup'] = _TRANSITION_TO_CLEANUP
DESCRIPTOR.message_types_by_name['state_changed'] = _STATE_CHANGED
DESCRIPTOR.message_types_by_name['state_busy'] = _STATE_BUSY
DESCRIPTOR.message_types_by_name['state_ready'] = _STATE_READY
DESCRIPTOR.message_types_by_name['log'] = _LOG

workflow = _reflection.GeneratedProtocolMessageType('workflow', (_message.Message,), dict(
  DESCRIPTOR = _WORKFLOW,
  __module__ = 'middleware_messages_pb2'
  # @@protoc_insertion_point(class_scope:workflow)
  ))
_sym_db.RegisterMessage(workflow)

workflow_request = _reflection.GeneratedProtocolMessageType('workflow_request', (_message.Message,), dict(
  DESCRIPTOR = _WORKFLOW_REQUEST,
  __module__ = 'middleware_messages_pb2'
  # @@protoc_insertion_point(class_scope:workflow_request)
  ))
_sym_db.RegisterMessage(workflow_request)

workflow_list = _reflection.GeneratedProtocolMessageType('workflow_list', (_message.Message,), dict(
  DESCRIPTOR = _WORKFLOW_LIST,
  __module__ = 'middleware_messages_pb2'
  # @@protoc_insertion_point(class_scope:workflow_list)
  ))
_sym_db.RegisterMessage(workflow_list)

workflow_load = _reflection.GeneratedProtocolMessageType('workflow_load', (_message.Message,), dict(
  DESCRIPTOR = _WORKFLOW_LOAD,
  __module__ = 'middleware_messages_pb2'
  # @@protoc_insertion_point(class_scope:workflow_load)
  ))
_sym_db.RegisterMessage(workflow_load)

workflow_list_request = _reflection.GeneratedProtocolMessageType('workflow_list_request', (_message.Message,), dict(
  DESCRIPTOR = _WORKFLOW_LIST_REQUEST,
  __module__ = 'middleware_messages_pb2'
  # @@protoc_insertion_point(class_scope:workflow_list_request)
  ))
_sym_db.RegisterMessage(workflow_list_request)

workflow_revert_state = _reflection.GeneratedProtocolMessageType('workflow_revert_state', (_message.Message,), dict(
  DESCRIPTOR = _WORKFLOW_REVERT_STATE,
  __module__ = 'middleware_messages_pb2'
  # @@protoc_insertion_point(class_scope:workflow_revert_state)
  ))
_sym_db.RegisterMessage(workflow_revert_state)

transition_to_state = _reflection.GeneratedProtocolMessageType('transition_to_state', (_message.Message,), dict(
  DESCRIPTOR = _TRANSITION_TO_STATE,
  __module__ = 'middleware_messages_pb2'
  # @@protoc_insertion_point(class_scope:transition_to_state)
  ))
_sym_db.RegisterMessage(transition_to_state)

transition_to_cleanup = _reflection.GeneratedProtocolMessageType('transition_to_cleanup', (_message.Message,), dict(
  DESCRIPTOR = _TRANSITION_TO_CLEANUP,
  __module__ = 'middleware_messages_pb2'
  # @@protoc_insertion_point(class_scope:transition_to_cleanup)
  ))
_sym_db.RegisterMessage(transition_to_cleanup)

state_changed = _reflection.GeneratedProtocolMessageType('state_changed', (_message.Message,), dict(
  DESCRIPTOR = _STATE_CHANGED,
  __module__ = 'middleware_messages_pb2'
  # @@protoc_insertion_point(class_scope:state_changed)
  ))
_sym_db.RegisterMessage(state_changed)

state_busy = _reflection.GeneratedProtocolMessageType('state_busy', (_message.Message,), dict(
  DESCRIPTOR = _STATE_BUSY,
  __module__ = 'middleware_messages_pb2'
  # @@protoc_insertion_point(class_scope:state_busy)
  ))
_sym_db.RegisterMessage(state_busy)

state_ready = _reflection.GeneratedProtocolMessageType('state_ready', (_message.Message,), dict(
  DESCRIPTOR = _STATE_READY,
  __module__ = 'middleware_messages_pb2'
  # @@protoc_insertion_point(class_scope:state_ready)
  ))
_sym_db.RegisterMessage(state_ready)

log = _reflection.GeneratedProtocolMessageType('log', (_message.Message,), dict(
  DESCRIPTOR = _LOG,
  __module__ = 'middleware_messages_pb2'
  # @@protoc_insertion_point(class_scope:log)
  ))
_sym_db.RegisterMessage(log)


# @@protoc_insertion_point(module_scope)



import argparse
import json
import logging
import sys
from pprint import pprint, pformat
from socket import *
import os

import np_logging 

logger = np_logging.getLogger(__name__)

R = {'mvr_request': ''}
encoding = 'utf-8'


class ResponseBuffer:
    def __init__(self):
        self.read_buffer = []

    def parse_buffer(self, buf):
        if not buf:
            return []
        buf = buf.decode()
        read_bracket_count = 0
        self.read_buffer.extend(buf)
        count = 0
        messages = []
        for i, c in enumerate(self.read_buffer):  # should maintain an internal pointer so it doesn't redo the list
            count += 1
            # update the "bracket stack"
            if c == '{':
                read_bracket_count += 1
            elif c == '}':
                read_bracket_count -= 1

            if read_bracket_count == 0:  # a full JSON string is available
                try:

                    messages.append(json.loads(''.join(self.read_buffer[i - count + 1: i + 1])))
                except TypeError:
                    logger.warning('%s | Error parsing message: %s', __class__.__name__, self.read_buffer)
                count = 0

        # strip prior json messages off the buffer
        if count == 0 and read_bracket_count == 0:  # There must be a better way
            self.read_buffer = []
        else:
            self.read_buffer = self.read_buffer[-count:]

        return messages


class MVRConnector:
    def __init__(self, args=None):
        self.response_buffer = ResponseBuffer()
        self.device_index_map = {}
        self._errors_since_last_success: int = 0
        self._delete_on_copy: bool = True
        self._recording: bool = False
        self._args = args
        self._mvr_sock = None
        self._host_to_camera_map = {}
        self._mvr_connected = False
        self.comp_ids = []  # temporary to allow for some debugging
        self.output_dir = "c$/ProgramData/AIBS_MPE/MVR/data/"
        try:
            self.connect_to_mvr()
        except Exception as err:
            logger.error(f'failed to connect to mvr:{err}')
            exit()

    def _recv(self):
        if not self._mvr_connected:
            self.connect_to_mvr()
            return

        try:
            ret_val = self._mvr_sock.recv(1024)
        except ConnectionResetError as e:
            logger.warning('%s | Connection reset error', __class__.__name__)
            self._mvr_connected = False
            return []
        except:
            return []
        return ret_val

    def read(self):
        buf = self._recv()
        return self.response_buffer.parse_buffer(buf)

    def _send(self, msg):
        """
        msg is a dictionary.
        _send creates json from the dictionary and sends it as a byte object
        """
        msg = json.dumps(msg).encode()
        logger.debug('%s | Sending: %s', __class__.__name__, msg)
        if not self._mvr_connected:
            self.connect_to_mvr()
        if not self._mvr_connected:
            return
        try:
            self._mvr_sock.send(msg)
        except (ConnectionResetError, ConnectionRefusedError):
            self._mvr_connected = False

    def connect_to_mvr(self):
        """
        Creates a STREAM Socket connection to the MultiVideoRecorder
        """
        self._mvr_sock = socket(AF_INET, SOCK_STREAM)
        self._mvr_sock.settimeout(10.0)
        if self._errors_since_last_success == 0:
            logger.debug('%s | Connecting on %s:%s', __class__.__name__, self._args["host"], self._args["port"])
        try:
            self._mvr_sock.connect((self._args['host'], self._args['port']))
            self._mvr_connected = True
        except OSError:
            if self._errors_since_last_success == 0:
                logger.debug('%s | Connection failed, will be re-attempted on the next write.', __class__.__name__)
            self._errors_since_last_success += 1
        else:
            self._errors_since_last_success = 0
            logger.debug('%s | Connection success: %s', __class__.__name__, self.read())

    def get_version(self):
        msg = {'mvr_request': 'get_version'}
        self._send(msg)
        return self.read()

    def start_display(self):
        msg = {'mvr_request': 'start_display'}
        self._send(msg)
        return self.read()

    def stop_display(self):
        msg = {'mvr_request': 'stop_display'}
        self._send(msg)
        return self.read()

    def start_record(self, file_name_prefix='', sub_folder='.', record_time=4800):
        self._send({'mvr_request': 'start_record',
                    'sub_folder': sub_folder,
                    'file_name_prefix': file_name_prefix,
                    'recording_time': record_time,
                    })

    def start_single_record(self, host, file_name_prefix='', sub_folder='.', record_time=4800):
        print(f'start single record on {host}')
        if host not in self._host_to_camera_map:
            comp = self.host_to_comp['host']
            logger.warning(f'Start Single Record: Can not find host {host} ({comp}) associated with a camera.')
            return

        self._send({"mvr_request": "start_record",
                    "camera_indices": [{"camera_index": f"Camera {self._host_to_camera_map[host]}"}],
                    "sub_folder": sub_folder,
                    "file_name_prefix": file_name_prefix,
                    "recording_time": record_time,
                    })

    def set_automated_ui(self, state):
        if state:
            message = {"mvr_request": "set_automated_ui"}

        else:
            message = {"mvr_request": "set_unautomated_ui"}

        self._send(message)
        return self.read()

    def stop_record(self):
        msg = {'mvr_request': 'stop_record'}
        self._send(msg)
        # return self.read()

    def stop_single_record(self, host):
        if host not in self._host_to_camera_map:
            comp = self.host_to_comp['host']
            logger.warning(f'Stop Single Record: Can not find host {host} ({comp}) associated with a camera.')
            return

        cam_id = self._host_to_camera_map[host]
        message = {"mvr_request": "stop_record",
                   "camera_indices": [f"Camera {cam_id}",
                                      ]
                   }
        self._send(message)

    def take_snapshot(self):
        self._send({"mvr_request": "take_snapshot"})

    def define_hosts(self, hosts):
        self._host_to_camera_map = {}
        for idx, host in enumerate(hosts):
            self._host_to_camera_map[host] = idx + 1

        # print(f'Host To Camera Map: {pformat(self._host_to_camera_map)}')
        self.host_to_comp = list(zip(hosts, self.comp_ids))

    def request_camera_ids(self):
        self._send({"mvr_request": "get_camera_ids"})
        return self.read()
    
    def highlight_camera(self, device_name):
        print(self.device_index_map)
        index = self.device_index_map[device_name]
        msg = {'mvr_request': 'toggle_highlight_camera',
               'camera': index
               }
        self._send(msg)

    def unhighlight_camera(self, panel):
        msg = {'mvr_request': 'unhighlight_panel'}
        self._send(msg)
        return self.read()

    def get_state(self):
        if self._recording:
            return 'BUSY', 'RECORDING'
        else:
            return 'READY', ''

    def shutdown(self):
        self._rep_sock.close()

    def _onclose(self):
        self.shutdown()

    @property
    def platform_info(self):
        return '0.1.0'





from np_workflows.workflows.shared import npxc

import np_logging

# def initialize_enter(state_globals) -> None: ... 
# def initialize_input(state_globals) -> None: ...
# def initialize_exit(state_globals) -> None: ...
def msg_enter(state_globals) -> None: 
    import pdb; pdb.set_trace()
    state_globals["external"]["transition_result"] = False # T/F just sets text at bottom green/red - switches to T in next state
    state_globals["external"]["status_message"] = 'status_message at bottom'
    state_globals["external"]["msg_text"] = 'msg_text in main widget'
    state_globals['external']['alert'] = True # T adds warning icon 
    # state_globals["external"]["next_state"] = 'warn' # sets next state by name
# def msg_input(state_globals) -> None: ... 
# def msg_exit(state_globals) -> None: ...
# def warn_enter(state_globals) -> None: ...
# def warn_input(state_globals) -> None: ...
# def workflow_complete_enter(state_globals) -> None: ...
# def workflow_complete_input(state_globals) -> None: ...
# def workflow_complete_exit(state_globals) -> None: ...


from __future__ import annotations

import contextlib
import datetime
import pickle
import functools
import pathlib
import typing
from typing import Any, Optional, Protocol, Union, Sequence, ClassVar, Type


from np_workflows.services import config
from np_workflows.services.protocols import Service, Startable, Verifiable, Stoppable, Finalizable
from np_workflows.services.proxies import ImageMVR, VideoMVR, Sync, NoCamstim, ScriptCamstim, SessionCamstim, JsonRecorder, NewScaleCoordinateRecorder
from np_workflows.services import open_ephys as OpenEphys
from np_workflows.experiments.baseclasses import Experiment, WithLims

import np_config
# import np_session
import np_logging

logger = np_logging.getLogger(__name__)


class NpUltra(Experiment):
        
    services: tuple[Service, ...] = (
        Sync, OpenEphys, ImageMVR, NewScaleCoordinateRecorder, NoCamstim,
    )
    "All services used in the experiment that support Service protocols."
    
    photodoc_services: tuple[Service, ...] = (
        ImageMVR, NewScaleCoordinateRecorder,
    )
    "Services called during photodoc capture."
    stimulus_services: tuple[Service, ...] = (
        NoCamstim,
    )
    
    def __init__(self, labtracks_mouse_id) -> None:    
        #TODO deepmerge config dicts 
        self.config : dict[str, dict[str, Any]]  = (
            np_config.from_zk('/projects/np_workflows/defaults/configuration')
            | np_config.from_zk(f'/projects/np_workflows/{self.__class__.__name__}/configuration')
            | np_config.from_zk(f'/rigs/NP.{self.rig_idx}')
        )
        self.session = f"{labtracks_mouse_id}_{datetime.datetime.now().strftime('%Y%m%d')}"
        OpenEphys.folder = self.session
        NoCamstim.remote_file = pathlib.Path('C:/Users/svc_neuropix/Desktop/run_blue_opto.bat')
    
    def advance_trial_idx(self, state_globals):
        if not hasattr(self, 'trial_idx'):
            self.trial_idx = 0
        else: 
            self.trial_idx += 1
        state_globals['external']['trial_idx'] = self.trial_idx
        OpenEphys.set_folder(f'{self.session}_{self.trial_idx}')
    
    @property
    def initial_services(self) -> tuple[Service, ...]:
        return tuple(set(self.services) - set(self.stimulus_services) - set(self.photodoc_services) ) 
    
    def start_recording(self):
        for service in self.initial_services:
            if isinstance(service, Startable):
                service.start()
        for service in self.initial_services:
            if isinstance(service, Verifiable):
                service.verify()
                
    def start_stimulus(self):
        for service in self.stimulus_services:
            if isinstance(service, Startable):
                service.start()
        for service in self.stimulus_services:
            if isinstance(service, Verifiable):
                service.verify()
                
    def stop_recording(self):
        for service in self.stimulus_services:
            if isinstance(service, Stoppable):
                service.stop()
        for service in self.services:
            if isinstance(service, Stoppable):
                service.stop()
                
        for service in self.stimulus_services:
            if isinstance(service, Finalizable):
                service.finalize()
        for service in self.services:
            if isinstance(service, Finalizable):
                service.finalize()
        
class Pretest(WithLims):
        
    services: tuple[Service, ...] = (
        Sync, OpenEphys, ImageMVR, VideoMVR, NewScaleCoordinateRecorder, ScriptCamstim,
    )
    "All services used in the experiment that support Service protocols."
    
    stim_services: tuple[Service, ...] = (
        NoCamstim,
    )
    
    photodoc_services: tuple[Service, ...] = (
        ImageMVR, NewScaleCoordinateRecorder,
    )
    "Services called during photodoc capture."
    
    recorder_services: tuple[Service, ...] = (
        # JsonRecorder, #! update with platform json recorder when available
    )
    
    def __init__(self,
        labtracks_mouse_id: int | str,
        lims_user_id: str,
        lims_session_id: Optional[int] = None,
        *args,
        **kwargs,
    ) -> None:
        
        logger.debug(f"Initializing {__class__.__name__}({labtracks_mouse_id}, {lims_user_id})")
        
        #TODO deepmerge config dicts 
        self.config : dict[str, dict[str, Any]]  = (
            np_config.from_zk('/projects/np_workflows/defaults/configuration') 
            | np_config.from_zk(f'/rigs/NP.{self.rig_idx}')
        )
        labtracks_mouse_id=self.config['pretest_mouse']
        logger.debug(f"{__class__.__name__} modifying mouse ID to use {labtracks_mouse_id} for pretest")
        super().__init__(labtracks_mouse_id, lims_user_id, lims_session_id, *args, **kwargs)
   
    @property
    def _project(self) -> str:
        "Accessed by WithLims.project getter."
        return f'Pretest_NP{self.rig_idx}'
    
    def apply_config_to_services(self):
        super().apply_config_to_services()
        OpenEphys.folder = self.folder


"""
Proxy classes for interacting with devices via zro/zmq.

Proxy class names must match the name of the proxy key in the config dict.
"""
from __future__ import annotations

import abc
import contextlib
import csv
import datetime
import json  # loading config from Sync proxy will instantiate datetime objects
import logging
import pathlib
import re
import time
from typing import Any, ClassVar, Literal, Mapping, Optional, Sequence

import np_logging
import np_config
import fabric 

from .config import Rig
from . import zro
from . import mvr
from . import utils
from .protocols import *

# TODO start proxies remotely if not connected

logger = np_logging.getLogger(__name__)

CONFIG = utils.config_from_zk()

ProxyState = tuple[Literal['', 'READY', 'BUSY'], str]

class Proxy(abc.ABC):
    
    # req proxy config - hardcode or overload ensure_config() 
    host: ClassVar[str]
    port: ClassVar[int]
    timeout: ClassVar[float]
    serialization: ClassVar[Literal['json', 'pickle']]
    
    # if a program needs to be launched (e.g. via RSC):
    rsc_app_id: str 
    
    # if device records:
    gb_per_hr: ClassVar[int | float]
    min_rec_hr: ClassVar[int | float]
    pretest_duration_sec: ClassVar[int | float]
    
    # for resulting data, if device records:
    data_root: ClassVar[Optional[pathlib.Path]] = None
    data_files: ClassVar[Optional[Sequence[pathlib.Path]]] = None
    
    # info
    exc: ClassVar[Optional[Exception]] = None
    
    latest_start: ClassVar[int] = 0
    "`time.time()` when the service was last started via `start()`."
    
    @classmethod
    def ensure_config(cls) -> None:
        """Updates any missing parameters for class proxy. 
        
        Is called in `get_proxy()` so any time we need the proxy, we have a
        correct config, without remembering to run `initialize()` or some such.
        """
        config = CONFIG.get(__class__.__name__, {}) # class where this function is defined 
        config.update(**CONFIG.get(cls.__name__, {})) # the calling class, if different

        # for proxy (reqd):
        if not hasattr(cls, 'host'):
            cls.host = config['host']
        if not hasattr(cls,'port'):
            cls.port = int(config['port'])
        if not hasattr(cls, 'timeout'):
            cls.timeout = float(config.get('timeout', 10.0))
        if not hasattr(cls, 'serialization'):    
            cls.serialization = config.get('serialization', 'json')
        
        # for pretest (reqd, not used if device doesn't record)
        if not hasattr(cls, 'pretest_duration_sec'):
            cls.pretest_duration_sec = config.get('pretest_duration_sec', 5)
        if not hasattr(cls, 'gb_per_hr'):
            cls.gb_per_hr = config.get('gb_per_hr', 2.0)
        if not hasattr(cls, 'min_rec_hr'):
            cls.min_rec_hr = config.get('min_rec_hr', 3.0)
            
        # for resulting data (optional):
        if not cls.data_root or cls.host not in cls.data_root.parts:
            relative_path = config.get('data', None)
            if relative_path:
                root = pathlib.Path(f"//{cls.host}/{relative_path}")
                try:
                    _ = root.exists()
                except OSError as exc:
                    cls.exc = exc
                    logger.exception("Error accessing %s data path: %s", cls.__name__, root)
                    raise FileNotFoundError(f"{cls.__name__} data path is not accessible: {root}") from exc
                else:
                    cls.data_root = root
        if hasattr(cls, 'data_root'):
            cls.data_root.mkdir(parents=True, exist_ok=True)
    
    @classmethod
    def launch(cls) -> None:
        utils.start_rsc_app(cls.host, cls.rsc_app_id)
        
        
    @classmethod
    def initialize(cls) -> None:
        cls.launch()
        with contextlib.suppress(AttributeError):
            del cls.proxy
        cls.proxy = cls.get_proxy()
        if isinstance(cls, Startable) and not cls.is_ready_to_start():
            if isinstance(cls, Finalizable):
                cls.finalize()
            if not cls.is_ready_to_start():
                logger.warning("%s not ready to start: %s", cls.__name__, cls.get_state())
                return
        if cls.data_root:
            cls.data_files = []
        cls.initialization = time.time()
        logger.info("%s initialized: ready for use", cls.__name__)

    @classmethod                      
    def test(cls) -> None:
        "Quickly verify service is working and ready for use, or raise `TestFailure`."
        logger.debug('Testing %s proxy', cls.__name__)
        if not cls.is_connected():
            raise TestFailure(f"{cls.__name__} not connected to {cls.host}:{cls.port}") from cls.exc
        logger.debug('%s proxy connection to %s:%s confirmed', cls.__name__, cls.host, cls.port)
        gb = cls.get_required_disk_gb()
        if not cls.is_disk_space_ok():
            raise TestFailure(f"{cls.__name__} free disk space on {cls.data_root.drive} doesn't meet minimum of {gb} GB") from cls.exc
        logger.debug('%s tested successfully', cls.__name__)     
    
    @classmethod
    def get_proxy(cls) -> zro.DeviceProxy:
        "Return a proxy to the service without re-creating unnecessarily."
        with contextlib.suppress(AttributeError):
            return cls.proxy
        cls.ensure_config()
        logger.debug('Creating %s proxy to %s:%s', cls.__name__, cls.host, cls.port)
        cls.proxy = zro.DeviceProxy(cls.host, cls.port, cls.timeout, cls.serialization)
        return cls.get_proxy()
               
    @classmethod
    def get_state(cls) -> ProxyState | dict:
        "Dict may be deprecated: is no longer returned by Sync or Camstim proxies."
        return cls.get_proxy().get_state()
    
    @classmethod
    def get_latest_data(cls: Recorder, glob:Optional[str]=None, subfolders:str='') -> list[pathlib.Path] | None:
        cls.ensure_config()
        if not cls.data_root:
            return None
        if subfolders == '/': # can alter path to drive root 
            subfolders = ''
        if not glob:
            glob = f'*{cls.raw_suffix}' if hasattr(cls, 'raw_suffix') else '*'
        if not hasattr(cls, 'latest_start'):
            return [max(utils.get_files_created_between(cls.data_root / subfolders, glob), key=lambda x: x.stat().st_mtime)]
        return utils.get_files_created_between(cls.data_root / subfolders, glob, cls.latest_start)
         
    @classmethod
    def get_required_disk_gb(cls) -> float:
        "Return the minimum disk space required prior to start (to .1 GB). Returns `0.0` if service generates no data."
        cls.ensure_config()
        if not isinstance(cls, Startable):
            return 0.0
        return round(cls.min_rec_hr * cls.gb_per_hr, 1)
    
    @classmethod
    def is_disk_space_ok(cls) -> bool:
        required = cls.get_required_disk_gb()
        if required == 0.0:
            return True
        try:
            free = utils.free_gb(cls.data_root)
        except FileNotFoundError as exc:
            cls.exc = exc
            logger.exception(f"{cls.__name__} data path not accessible: {cls.data_root}")
            return False
        else:
            logger.debug('%s free disk space on %s: %s GB', cls.__name__, cls.data_root.drive, free)
            return free > required
            
    @classmethod
    def is_connected(cls) -> bool:
        if not utils.is_online(cls.host):
            cls.exc = ConnectionError(f"No response from {cls.host}: may be offline or unreachable")
            return False
        try:
            _ = cls.get_proxy().uptime
        except zro.ZroError as exc:
            cls.exc = exc
            logger.exception(f"{cls.__name__} proxy connection to {cls.host}:{cls.port} failed")
            return False
        try:
            _ = cls.get_state()
        except zro.ZroError as exc:
            cls.exc = exc
            logger.exception(f"{cls.__name__} proxy connection to {cls.host}:{cls.port} failed")
            return False
        return True


class CamstimSyncShared(Proxy):
    
    started_state: ClassVar[Sequence[str]]
    
    @classmethod
    def is_ready_to_start(cls) -> bool:
        if cls.is_started():
            return False
        state = cls.get_state()
        if isinstance(state, Mapping) and state.get('message', '') == 'READY':
            return True
        if isinstance(state, Sequence) and 'READY' in state:
            return True
        return False
        
    @classmethod
    def is_started(cls) -> bool:
        return len(state := cls.get_state()) and all(msg in state for msg in cls.started_state)
        
    @classmethod
    def start(cls) -> None:
        logger.debug("Starting %s", cls.__name__)
        if cls.is_started():
            logger.warning("%s already started - should be stopped manually", cls.__name__)
            return
            # otherwise, Sync - for example - would stop current recording and start another
        if not cls.is_ready_to_start():
            logger.error("%s not ready to start: %s", cls.__name__, cls.get_state())
            raise AssertionError(f"{cls.__name__} not ready to start: {cls.get_state()}")
        cls.latest_start = time.time()
        cls.get_proxy().start()
    
    @classmethod
    def pretest(cls) -> None:
        "Test all critical functions"
        with utils.debug_logging():
            logger.debug('Starting %s pretest', cls.__name__)
            cls.initialize() # calls test()

            with utils.stop_on_error(cls):
                cls.start()
                time.sleep(1)
                cls.verify()
                time.sleep(cls.pretest_duration_sec)
                # stop() called by context manager at exit, regardless
            cls.finalize()
            cls.validate()
        logger.info("%s pretest complete", cls.__name__)
                       
    @classmethod
    def verify(cls) -> None:
        "Assert latest data file is currently increasing in size, or raise AssertionError."
        if not cls.is_started():
            logger.warning("Cannot verify %s if not started: %s", cls.__name__, cls.get_state())
            raise AssertionError(f"{cls.__name__} not started: {cls.get_state()}")
        if cls.data_root and not utils.is_file_growing(cls.get_latest_data()[-1]):
            raise AssertionError(f"{cls.__name__} latest data file is not increasing in size: {cls.get_latest_data()[-1]}")
        logger.debug("%s latest data file is increasing in size", cls.__name__)
    
    @classmethod
    def stop(cls) -> None:
        logger.debug("Stopping %s", cls.__name__)
        cls.get_proxy().stop()
    
    # --- End of possible Camstim/Sync shared methods ---
 
    # --- Sync-specific methods ---
class Sync(CamstimSyncShared):
    host = Rig.Sync.host
    started_state = ("BUSY", "RECORDING")
    raw_suffix: ClassVar[int | float] = '.sync'
    rsc_app_id: str = 'sync_device'
    
    @classmethod
    def ensure_config(cls) -> None:
        """Updates any missing parameters for class proxy. 
        
        Is called in `get_proxy()` so any time we need the proxy, we have a
        correct config, without remembering to run `initialize()` or some such.
        """
        config = CONFIG.get(__class__.__name__, {}) # class where this function is defined 
        config.update(**CONFIG.get(cls.__name__, {})) # the calling class, if different

        # for proxy (reqd):
        if not hasattr(cls, 'host'):
            cls.host = config['host']
        if not hasattr(cls,'port'):
            cls.port = int(config['port'])
        if not hasattr(cls, 'timeout'):
            cls.timeout = float(config.get('timeout', 10.0))
        if not hasattr(cls, 'serialization'):    
            cls.serialization = config.get('serialization', 'json')
        
        # for pretest (reqd, not used if device doesn't record)
        if not hasattr(cls, 'pretest_duration_sec'):
            cls.pretest_duration_sec = config.get('pretest_duration_sec', 5)
        if not hasattr(cls, 'gb_per_hr'):
            cls.gb_per_hr = config.get('gb_per_hr', 2.0)
        if not hasattr(cls, 'min_rec_hr'):
            cls.min_rec_hr = config.get('min_rec_hr', 3.0)
            
        # for resulting data (optional):
        if not cls.data_root or cls.host not in cls.data_root.parts:
            relative_path = config.get('data', None)
            if relative_path:
                root = pathlib.Path(f"//{cls.host}/{relative_path}")
                try:
                    _ = root.exists()
                except OSError as exc:
                    cls.exc = exc
                    logger.exception("Error accessing %s data path: %s", cls.__name__, root)
                    raise FileNotFoundError(f"{cls.__name__} data path is not accessible: {root}") from exc
                else:
                    cls.data_root = root
        if hasattr(cls, 'data_root'):
            cls.data_root.mkdir(parents=True, exist_ok=True)
    
    @classmethod
    def finalize(cls) -> None:
        logger.debug('Finalizing %s', cls.__name__)
        if cls.is_started():
            cls.stop()
        while not cls.is_ready_to_start():
            logger.debug("Waiting for %s to finish processing", cls.__name__)
            time.sleep(1) # TODO add backoff module
        if not cls.data_files:
            cls.data_files = []
        cls.data_files.extend(new := cls.get_latest_data('*.h5'))
        logger.debug("%s processing finished: %s", cls.__name__, [_.name for _ in new])
        
    @classmethod
    def shutdown(cls) -> None:
        logger.debug("Shutting down %s", cls.__name__)
        cls.stop()
        try:
            del cls.proxy
        except Exception as exc:
            logger.debug("Failed to delete %s proxy: %s", cls.__name__, exc)
            cls.exc = exc        
            
    @classmethod
    def get_config(cls) -> dict[str, Any | datetime.datetime]:
        "Sync config, including `line_labels` and `frequency`"
        if cls.serialization in ("json", "j"):
            return eval(cls.get_proxy().config)
        if cls.serialization in ("pickle", "pkl", "p"):
            return cls.get_proxy().config
    
    @classmethod
    def validate(cls, data: Optional[pathlib.Path] = None) -> None:
        "Check that data file is valid, or raise AssertionError."
        logger.debug('Validating %s data', cls.__name__)
        if not data and bool(files := cls.get_latest_data('*.h5')):
            data = files[-1]
            logger.debug('No data file provided: validating most-recent data in %s: %s', cls.data_root, data.name)
            if cls.is_started():
                logger.warning(f"Attempted to validate current data file while recording")
                return
            elif not cls.is_ready_to_start():
                cls.finalize()
        try:
            import h5py
        except ImportError:
            logger.warning('h5py not installed: cannot open Sync data')
            cls.min_validation(data)
        else:
            cls.full_validation(data)
    
    @classmethod
    def full_validation(cls, data: pathlib.Path) -> None:
        line_labels: dict = cls.get_config()['line_labels']
        # TODO 
        pass
        
    @classmethod
    def min_validation(cls, data: pathlib.Path) -> None:
        if data.stat().st_size == 0:
            raise AssertionError(f"Empty file: {data}")
        if data.suffix != '.h5':
            raise FileNotFoundError(f"Expected .sync to be converted to .h5 immediately after recording stopped: {data}")
        logger.debug('%s minimal validation passed for %s', cls.__name__, data.name)

class Phidget(CamstimSyncShared):
    host = Rig.Stim.host
    rsc_app_id = 'phidget_server'
    
class Camstim(CamstimSyncShared):
    
    host = Rig.Stim.host
    started_state = ("BUSY", "Script in progress.")
    rsc_app_id = 'camstim_agent'
    
    @classmethod
    def launch(cls) -> None:
        super().launch()
        Phidget.launch()
        
    @classmethod
    def get_config(cls) -> dict[str, Any]:
        return cls.get_proxy().config
    
    @classmethod
    def ensure_config(cls) -> None:
        """Updates any missing parameters for class proxy. 
        
        Is called in `get_proxy()` so any time we need the proxy, we have a
        correct config, without remembering to run `initialize()` or some such.
        """
        config = CONFIG.get(__class__.__name__, {}) # class where this function is defined 
        config.update(**CONFIG.get(cls.__name__, {})) # the calling class, if different

        # for proxy (reqd):
        if not hasattr(cls, 'host'):
            cls.host = config['host']
        if not hasattr(cls,'port'):
            cls.port = int(config['port'])
        if not hasattr(cls, 'timeout'):
            cls.timeout = float(config.get('timeout', 10.0))
        if not hasattr(cls, 'serialization'):    
            cls.serialization = config.get('serialization', 'json')
        
        # for pretest (reqd, not used if device doesn't record)
        if not hasattr(cls, 'pretest_duration_sec'):
            cls.pretest_duration_sec = config.get('pretest_duration_sec', 5)
        if not hasattr(cls, 'gb_per_hr'):
            cls.gb_per_hr = config.get('gb_per_hr', 2.0)
        if not hasattr(cls, 'min_rec_hr'):
            cls.min_rec_hr = config.get('min_rec_hr', 3.0)
            
        # for resulting data (optional):
        if not cls.data_root or cls.host not in cls.data_root.parts:
            relative_path = config.get('data', None)
            if relative_path:
                root = pathlib.Path(f"//{cls.host}/{relative_path}")
                try:
                    _ = root.exists()
                except OSError as exc:
                    cls.exc = exc
                    logger.exception("Error accessing %s data path: %s", cls.__name__, root)
                    raise FileNotFoundError(f"{cls.__name__} data path is not accessible: {root}") from exc
                else:
                    cls.data_root = root
        if hasattr(cls, 'data_root') and cls.data_root is not None:
            cls.data_root.mkdir(parents=True, exist_ok=True)
    
    @classmethod
    def finalize(cls) -> None:
        logger.info('Finalizing %s', cls.__name__)
        if cls.is_started():
            cls.stop()
        while not cls.is_ready_to_start():
            logger.debug("Waiting for %s to finish processing", cls.__name__)
            time.sleep(1) # TODO add backoff module
        if not cls.data_files:
            cls.data_files = []
        cls.data_files.extend(new := cls.get_latest_data('*.pkl'))
        logger.info("%s added new data: %s", cls.__name__, [_.name for _ in new])
        
class ScriptCamstim(Camstim):
    script: ClassVar[str]
    "path to script on Stim computer"
    params: ClassVar[dict[str, Any]] = {}
    
    @classmethod
    def pretest(cls) -> None:
        cls.script = 'C:/ProgramData/StimulusFiles/dev/bi_script_pretest_v2.py'
        super().pretest()
        
    @classmethod
    def start(cls):
        cls.latest_start = time.time()
        cls.get_proxy().start_script(cls.script, cls.params)
        
class SessionCamstim(Camstim):
    lims_user_id: ClassVar[str]
    labtracks_mouse_id: ClassVar[int]
    @classmethod
    def start(cls):
        cls.latest_start = time.time()
        cls.get_proxy().start_session(cls.labtracks_mouse_id, cls.lims_user_id)#, cls.params)
    
    @classmethod
    def pretest(cls) -> None:
        cls.labtracks_mouse_id = 598796
        cls.lims_user_id = 'ben.hardcastle'
        super().pretest()
    
class NoCamstim(Camstim):
    "Run remote files (e.g. .bat) without sending directly to Camstim Agent"
    
    remote_file: ClassVar[str | pathlib.Path]
    extra_args: ClassVar[list[str]] = []
    ssh: ClassVar[fabric.Connection]
    user: ClassVar[str] = 'svc_neuropix'
    password: ClassVar[str]
    
    # @classmethod
    # def pretest(cls) -> None:
    #     cls.remote_file = 
    #     super().pretest()
        
    @classmethod
    def get_ssh(cls) -> fabric.Connection:
        with contextlib.suppress(AttributeError):
            return cls.ssh
        cls.initialize()
        return cls.ssh  

    @classmethod
    def initialize(cls) -> None:
        if not hasattr(cls, 'password'):
            cls.password = input(f"{cls.__name__} | Enter password for {cls.host}: ")
        cls.remote_file = utils.unc_to_local(pathlib.Path(cls.remote_file))
        cls.ssh = fabric.Connection(cls.host, cls.user, connect_kwargs=dict(password=cls.password))
        super().initialize()
        cls.test()
        
    @classmethod
    def test(cls) -> None:
        super().test()
        logger.debug(f"{cls.__name__} | Testing")
        try:
            result = cls.get_ssh().run('hostname', hide=True)
        except Exception as exc:
            raise TestFailure(f"{cls.__name__} Error connecting to {cls.host} via ssh: {exc!r}. Is this password correct? {cls.password}")
        else:
            if result.exited != 0:
                raise TestFailure(f"{cls.__name__} Error connecting to {cls.host} via ssh: {result}")
            logger.debug(f"{cls.__name__} | Connected to {cls.host} via ssh")
        
        try:
            result = cls.get_ssh().run(f'type {cls.remote_file}', hide=True)
        except Exception as exc:
            extra = f" | '{exc.result.command}': {exc.result.stderr.strip()!r}" if hasattr(exc, 'result') else ""
            raise TestFailure(f"{cls.__name__} | Error calling ssh-executed command{extra}")
        else:
            if result.exited != 0:
                raise TestFailure(f"{cls.__name__} Error accessing {cls.remote_file} on {cls.host} - is filepath correct? {result}")
            logger.debug(f"{cls.__name__} | {cls.remote_file} is accessible via ssh on {cls.host}")
        
    @classmethod
    def start(cls):
        
        if cls.is_started():
            logger.warning(f"{cls.__name__} already started")
            return
        logger.debug(f"{cls.__name__} | Starting {cls.remote_file} on {cls.host}")
        cls.latest_start = time.time()
        cls.get_ssh().run(f'call {cls.remote_file} {cls.extra_args}')
    
    @classmethod
    def verify(cls):
        logger.warning(f"{cls.__name__} | No verification implemented")
    
class MouseDirector(CamstimSyncShared):
    host = Rig.Mon.host
    gb_per_hr = 0
    serialization = "json"
    
class Cam3d(CamstimSyncShared):
    host = Rig.Mon.host
    serialization = 'json'
    started_state = ("READY", "CAMERAS_OPEN,CAMERAS_ACQUIRING")
    
    @classmethod
    def initialize(cls) -> None:
        cls.get_proxy().enable_cameras()


class MVR(CamstimSyncShared):
    # req proxy config - hardcode or overload ensure_config() 
    host: ClassVar[str] = Rig.Mon.host
    port: ClassVar[int]
    
    re_aux: re.Pattern = re.compile('aux|USB!', re.IGNORECASE)
    
    @classmethod
    def is_connected(cls) -> bool:
        if not utils.is_online(cls.host):
            cls.exc = ConnectionError(f"No response from {cls.host}: may be offline or unreachable")
            return False
        if not cls.get_proxy()._mvr_connected:
            cls.exc = ConnectionError(f"MVR likely not running on {cls.host}")
            return False
        try:
            _ = cls.get_camera_status()
        except ConnectionError as exc:
            cls.exc = exc
            return False
        return True
    

    @classmethod
    def initialize(cls) -> None:
        with contextlib.suppress(AttributeError):
            del cls.proxy
        cls.proxy = cls.get_proxy()
        cls.test()
        cls.configure_cameras()
        _ = cls.get_proxy().read() # empty buffer
        if isinstance(cls, Startable) and not cls.is_ready_to_start():
            if cls.is_started() and isinstance(cls, Stoppable):
                cls.stop()
            if isinstance(cls, Finalizable):
                cls.finalize()
            if not cls.is_ready_to_start():
                logger.warning("%s not ready to start: %s", cls.__name__, cls.get_state())
                return
        if cls.data_root:
            cls.data_files = []
        cls.initialization = time.time()
        logger.info("%s initialized: ready for use", cls.__name__)
        
    @classmethod
    def shutdown(cls) -> None:
        cls.get_proxy()._mvr_sock.close()  
        del cls.proxy
        
    @classmethod
    def get_proxy(cls) -> mvr.MVRConnector:
        with contextlib.suppress(AttributeError):
            return cls.proxy
        cls.ensure_config()
        logger.debug('Creating %s proxy to %s:%s', cls.__name__, cls.host, cls.port)
        cls.proxy = mvr.MVRConnector({'host':cls.host, 'port':cls.port})
        cls.proxy._mvr_sock.settimeout(cls.timeout)
        return cls.get_proxy()
    
    @classmethod
    def get_cameras(cls) -> list[dict[str, str]]:
        if not hasattr(cls, 'all_cameras'):
            cls.get_proxy().read()
            cls.all_cameras = cls.get_proxy().request_camera_ids()[0]['value']
        return cls.all_cameras
    
    @classmethod
    def get_camera_status(cls) -> list[dict[str, str]]:
        _ = cls.get_proxy().read() # empty buffer
        _ = cls.get_proxy()._send({"mvr_request":"get_camera_status"})
        for msg in reversed(cls.get_proxy().read()):
            if msg.get('mvr_response', '') == 'get_camera_status' and (cams := msg.get('value', [])):
                break
        else:
            logger.error("Could not get camera status from %s", cls.host)
            raise ConnectionError(f"Could not get camera status from {cls.host}")
        return [_ for _ in cams if any(_['camera_id'] == __['id'] for __ in cls.get_cameras())]
    
    @classmethod
    def get_state(cls) -> ProxyState:
        if not cls.is_connected():
            return '', 'MVR_CLOSED'
        status = cls.get_camera_status()
        # cam status could change between calls, so only get once  
        if any(not _['is_open'] for _ in status):
            return '', 'CAMERA_CLOSED'
        if any(not _['is_streaming'] for _ in status):
            return '', 'CAMERA_NOT_STREAMING'
        if cls.get_cameras_recording(status):
            return 'BUSY', 'RECORDING'
        return 'READY', ''
        
    @classmethod
    def get_cameras_recording(cls, status=None) -> list[dict[str, str]]:
        return [_ for _ in status or cls.get_camera_status() if _['is_recording']]
    
    @classmethod
    def is_ready_to_start(cls) -> bool:
        if cls.is_started():
            return False
        return all(_['is_open'] and _['is_streaming'] and not _['is_recording'] for _ in cls.get_camera_status())
    
    @classmethod
    def configure_cameras(cls) -> None:
        "Set MVR to record video from subset of all cameras, via `get_cameras` (implemented by subclass)"
        cam_ids = [_['id'] for _ in cls.get_cameras()]
        cls.get_proxy().define_hosts(cam_ids)
        cls.get_proxy().start_display()
    
    
class ImageMVR(MVR):
    gb_per_hr: ClassVar[int | float]
    min_rec_hr: ClassVar[int | float]
    label: ClassVar[str]
    "Rename file after capture to include label"
    
    # TODO ready state is if Aux cam is_open
    @classmethod
    def get_cameras(cls) -> list[dict[str, str]]:
        "Aux cam only"
        cams = super().get_cameras()
        return [_ for _ in cams if cls.re_aux.search(_['label'])]
        
    @classmethod
    def start(cls):
        if not cls.is_ready_to_start():
            # TODO display state, wait on user input to continue
            logger.error("%s not ready to start: %s", cls.__name__, cls.get_state())
            raise AssertionError(f"{cls.__name__} not ready to start: {cls.get_state()}")
        cls.latest_start = time.time()
        cls.get_proxy().take_snapshot()
        
    @classmethod
    def stop(cls):
        "Overload parent method to do nothing"
        pass
    
    @classmethod
    def is_started(cls) -> bool:
        for msg in cls.get_proxy().read():
            if msg.get('mvr_broadcast', '') == 'snapshot_converted':
                return True 
            if msg.get('mvr_broadcast', '') == 'snapshot_failed':
                return False
        return False
    
    @classmethod
    def verify(cls):
        "Overload parent method to do nothing"
        pass
    
    # TODO
    @classmethod
    def validate(cls) -> None: 
        logger.warning('%s.validate() not implemented', cls.__name__)
    
    @classmethod
    def finalize(cls) -> None:
        logger.debug('Finalizing %s', cls.__name__)
        t0 = time.time()
        timedout = lambda: time.time() > t0 + 10
        while (
            (
                cls.is_started() 
                or not cls.is_ready_to_start()
                or not cls.get_latest_data('*')
                or cls.get_latest_data('.bmp')
            ) 
            and not timedout()
            ):
            logger.debug("Waiting for %s to finish processing", cls.__name__)
            time.sleep(1) # TODO add backoff module
        if timedout():
            logger.warning("Timed out waiting for %s to finish processing", cls.__name__)
            return
        if not hasattr(cls, 'data_files') or not cls.data_files:
            cls.data_files = []
        new = cls.get_latest_data('*')
        if hasattr(cls, 'label') and cls.label:
            new = [_.rename(_.with_stem(f"{_.stem}_{cls.label}")) for _ in new]
        cls.data_files.extend(new)
        logger.debug("%s processing finished: %s", cls.__name__, [_.name for _ in new])
            
class VideoMVR(MVR):
    pretest_duration_sec: ClassVar[int | float]
    gb_per_hr: ClassVar[int | float]
    min_rec_hr: ClassVar[int | float]
    
    raw_suffix: ClassVar[str] = '.mp4'
    
    started_state = ('BUSY', 'RECORDING')
    
    @classmethod
    def get_cameras(cls) -> list[dict[str, str]]:
        "All available cams except Aux"
        cams = super().get_cameras()
        return [_ for _ in cams if cls.re_aux.search(_['label']) is None]

    @classmethod
    def start(cls) -> None:
        cls.latest_start = time.time()
        cls.get_proxy().start_record(
            record_time=365 * 60, # sec
        )

    @classmethod
    def verify(cls) -> None:
        "Assert data exists since latest start, or raise AssertionError."
        # files grow infrequently while MVR's recording - checking their size
        # is unreliable
        if not cls.is_started():
            logger.warning("Cannot verify %s if not started: %s", cls.__name__, cls.get_state())
            raise AssertionError(f"{cls.__name__} not started: {cls.get_state()}")
        if datetime.datetime.fromtimestamp(cls.latest_start) < datetime.datetime.now() - datetime.timedelta(seconds=cls.pretest_duration_sec):
            time.sleep(cls.pretest_duration_sec)
        if not (files := cls.get_latest_data()) or len(files) < len(cls.get_cameras_recording()):
            raise AssertionError(f"{cls.__name__} files do not match the number of cameras: {files}")
        logger.debug("%s verified: %s cameras recording to disk", cls.__name__, len(files))
    
    @classmethod
    def stop(cls) -> None:
        cls.get_proxy().stop_record()

    @classmethod
    def is_started(cls) -> bool:       
        if len(state := cls.get_state()) and all(msg in state for msg in cls.started_state):
            return True
        return False 
    
    @classmethod
    def finalize(cls) -> None:
        logger.debug('Finalizing %s', cls.__name__)
        if cls.is_started():
            cls.stop()
        t0 = time.time()
        timedout = lambda: time.time() > t0 + 30
        while not cls.is_ready_to_start() and not timedout():
            logger.debug("Waiting for %s to finish processing", cls.__name__)
            time.sleep(1) # TODO add backoff module
        if timedout:
            logger.warning("Timed out waiting for %s to finish processing", cls.__name__)
            return
        if not hasattr(cls, 'data_files'):
            cls.data_files = []
        cls.data_files.extend(new := (cls.get_latest_data('*.mp4') + cls.get_latest_data('*.json')))
        logger.debug("%s processing finished: %s", cls.__name__, [_.name for _ in new])
        
    @classmethod
    def validate(cls) -> None:
        logger.warning('%s.validate() not implemented', cls.__name__)

class JsonRecorder:
    "Just needs a `start` method that calls `write()`."
    
    log_name: ClassVar[str] 
    log_root: ClassVar[pathlib.Path]
    
    @classmethod
    def pretest(cls) -> None:
        cls.initialize()
        cls.start()
        cls.validate()
        
    @classmethod
    def ensure_config(cls)  -> None:
                
        config = CONFIG.get(__class__.__name__, {}) # class where this function is defined 
        config.update(**CONFIG.get(cls.__name__, {})) # the calling class, if different

        if not hasattr(cls, 'log_name'):
            cls.log_name = config.get('log_name', '{}_.json')
        cls.log_name = cls.log_name.format(datetime.datetime.now().strftime("%Y-%m-%d_%H%M%S"))
            
        if not hasattr(cls, 'log_root'):
            cls.log_root = config.get('log_root', '.')
        cls.log_root = pathlib.Path(cls.log_root).resolve()
               
    @classmethod
    def initialize(cls) -> None:
        logger.debug('%s initializing', __class__.__name__)
        cls.ensure_config()
        cls.initialization = time.time()
        log = (cls.log_root / cls.log_name).with_suffix('.json')
        log.parent.mkdir(parents=True, exist_ok=True)
        log.touch()
        cls.all_files = [log]
        cls.test()
        
    @classmethod
    def test(cls) -> None:
        logger.debug('%s testing', __class__.__name__)
        try:
            _ = cls.get_current_log().open('w')
        except OSError as exc:
            raise TestFailure(f'{__class__.__name__} failed to open {cls.get_current_log()}') from exc
        
    @classmethod
    def get_current_log(cls) -> pathlib.Path:
        if not hasattr(cls, 'initialization'):
            cls.initialize()
        return cls.all_files[-1]

    @classmethod
    def read(cls) -> dict[str, str|float]:
        try:
            data = json.loads(cls.get_current_log().read_bytes())
        except json.JSONDecodeError:
            if cls.get_current_log().stat().st_size:
                raise
            data = {} # file was empty
        else:
            logger.debug('%s read from %s', __class__.__name__, cls.get_current_log())
        return data
    
    @classmethod
    def write(cls, value: dict) -> None:
        try:
            data = cls.read()
        except json.JSONDecodeError:
            data = {}
            file = cls.get_current_log().with_suffix('.new.json')
            file.touch()
            cls.all_files.append(file)
        else:
            file = cls.get_current_log()
        data.update(value)
        file.write_text(json.dumps(data, indent=4, sort_keys=False, default=str))
        logger.debug('%s wrote to %s', __class__.__name__, file)
     
    @classmethod
    def validate(cls) -> None:
        if not (log := cls.read()):
            cls.exc = TestFailure(f'{cls.__name__} failed to validate because log is empty: {cls.get_current_log()}')
            logger.error('%s failed to validate: log is empty %s', cls.__name__, cls.get_current_log(), exc_info=cls.exc)
        logger.debug('%s validated', __class__.__name__)
        
class YamlRecorder(JsonRecorder):
    
    @classmethod
    def test(cls) -> None:
        logger.debug('%s testing', __class__.__name__)
        super().test()
        try:
            import yaml
        except ImportError as exc:
            raise TestFailure(f'{__class__.__name__} failed to import yaml') from exc
        
    @classmethod
    def finalize(cls) -> None:
        logger.debug('Finalizing %s', __class__.__name__)
        log = json.load(cls.get_current_log().read_bytes())
        with utils.suppress(AttributeError, OSError): # if this fails we still have the json file
            yaml.dump(log, cls.get_current_log().with_suffix('.yaml').read_bytes())
        
class NewScaleCoordinateRecorder(JsonRecorder):
    "Gets current manipulator coordinates and stores them in a file with a timestamp."
    
    host: ClassVar[str] = Rig.Mon.host
    data_root: ClassVar[pathlib.Path]
    data_name: ClassVar[str] = 'log.csv'
    data_fieldnames: ClassVar[Sequence[str]] = ('last_moved', 'manipulator', 'x', 'y', 'z', 'x_virtual', 'y_virtual', 'z_virtual')
                
    num_probes: ClassVar[int] = 6
    max_travel: ClassVar[float]
    log_name: ClassVar[str] = 'newscale_coords_{}.json'
    log_root: ClassVar[pathlib.Path] = pathlib.Path('.').resolve() #! move to config after testing
    label: ClassVar[str] = '' 
    "A label to tag each entry with"
    latest_start: ClassVar[int] = 0
    "`time.time()` when the service was last started via `start()`."
    
    @classmethod
    def get_current_data(cls) -> pathlib.Path:
        cls.ensure_config()
        return cls.data_root / cls.data_name
    
    @classmethod
    def get_coordinates(cls) -> dict[str, float]:
        with cls.get_current_data().open('r') as _:
            reader = csv.DictReader(_, fieldnames=cls.data_fieldnames)
            rows = list(reader)
        coords = {}
        for row in reversed(rows): # search for the most recent coordinates
            if len(coords.keys()) == cls.num_probes: 
                break # we have an entry for each probe
            if (m := row.pop(cls.data_fieldnames[1]).strip()) not in coords:
                coords[m] = {}
                for k, v in row.items():
                    if 'virtual' in k:
                        continue
                    v = v.strip()
                    with contextlib.suppress(ValueError):
                        v = float(v)
                    coords[m].update({k: v})
        coords['label'] = cls.label
        logger.debug('%s retrieved coordinates: %s', cls.__name__, coords)
        return coords
    
    @classmethod
    def start(cls):
        cls.latest_start = time.time()
        cls.write({str(datetime.datetime.now()): cls.get_coordinates()})
        
    @classmethod
    def test(cls) -> None:
        super().test()
        logger.debug('%s testing', __class__.__name__)
        try:
            _ = cls.get_current_data().open('r')
        except OSError as exc:
            raise TestFailure(f'{cls.__name__} failed to open {cls.get_current_data()}') from exc
        try:
            _ = cls.get_coordinates()
        except Exception as exc:
            raise TestFailure(f'{cls.__name__} failed to get coordinates') from exc
        else:
            logger.info('%s test passed', cls.__name__)

    @classmethod
    def ensure_config(cls)  -> None:
        
        super().ensure_config()
        
        if CONFIG.get('services', {}):
            config = CONFIG['services'].get(__class__.__name__, {})
            config.update(**CONFIG['services'].get(cls.__name__, {}))
        else:
            config = CONFIG.get(__class__.__name__, {}) # class where this function is defined 
            config.update(**CONFIG.get(cls.__name__, {})) # the calling class, if different

        if not hasattr(cls, 'host'):
            cls.host = config['host']
            
        # for resulting data
        if not hasattr(cls, 'data_root') or cls.host not in pathlib.Path(cls.data_root).parts:
            relative_path = config['data']
            if relative_path:
                root = pathlib.Path(f"//{cls.host}/{relative_path}")
                try:
                    _ = root.exists()
                except OSError as exc:
                    cls.exc = exc
                    logger.exception("Error accessing %s data path: %s", cls.__name__, root)
                    raise FileNotFoundError(f"{cls.__name__} data path is not accessible: {root}") from exc
                else:
                    cls.data_root = root
                    
        if not hasattr(cls, 'data_name'):
            cls.data_name = config['data_name']
        if not hasattr(cls, 'data_fieldnames'):
            cls.data_fieldnames = config['data_fieldnames']
        
